//---------------------------------------------------------------------------
// topology 2-dimensional
//---------------------------------------------------------------------------
//                              Sum
//       0 -> 4 -> 8  -> 12      24
//       |    |    |     |
//       1 -> 5 -> 9  -> 13      28
//       |    |    |     |
//       2 -> 6 -> 10 -> 14      32
//       |    |    |     |
//       3 -> 7 -> 11 -> 15      36
//
//  Sum  6    22   38    54
//
#if (defined __linux__) || (defined _AIX) || (defined __APPLE__)
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <unistd.h>
#elif (defined _WIN32) || (defined _WIN64)
	#include <conio.h>
	#include <direct.h>
#endif

#include<mpi.h>
#include<stdlib.h>
#include<stdio.h>
#include <iostream>
using namespace std;

int main(int argc, char** argv)
{
	int i, rank, size, src, dest, direction = 0, mess1 = 0, mess2 = 0, mysumm = 0;
	int dims[2], periods[2];
	double time;
	MPI_Request request;
	MPI_Status status;
	MPI_Comm mycomm_cart;

	MPI_Init(&argc, &argv);
	MPI_Barrier(MPI_COMM_WORLD);
	time = -MPI_Wtime();

	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	if (rank == 0) {
		printf("\nThere are %d processes.", size);
		fflush(stdout);
	}

	if (size != 16) {
		if (rank == 0)
			cout << "\n\nPlease, set the number of processes to 16!\n"; // size = 16 !!!
		goto end;
	}

	dims[0] = size / 4;
	dims[1] = 4;
	periods[0] = 1;
	periods[1] = 1;

	MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, true, &mycomm_cart);
	mess1 = rank;

	if (rank == 0) {
		cout << "\n---------------------------------------------------------------";
		cout << "\nSelect a direction:\n  0. horizontal\n"
			<< "  1. vertical\n\n  direction = ";
		cin >> direction;
		cout << "---------------------------------------------------------------\n";
		fflush(stdout);
	}
	MPI_Bcast(&direction, 1, MPI_INT, 0, MPI_COMM_WORLD);

	for (i = 0; i < size / 4; i++) {
		MPI_Cart_shift(mycomm_cart, direction, 1, &src, &dest);
		printf("%d -> %d", src, rank);
		fflush(stdout);
		MPI_Issend(&mess1, 1, MPI_INT, dest, 0, mycomm_cart, &request);
		MPI_Recv(&mess2, 1, MPI_INT, src, 0, mycomm_cart, &status);
		MPI_Wait(&request, &status);
		mysumm += mess2;
		mess1 = mess2;
	}

	printf("process %d: summ of messages = %d", rank, mysumm);

end:
	time += MPI_Wtime();
	if (rank == 0)
		printf("\nTime = %10.6f s\n", time);

	fflush(stdout);
	MPI_Finalize();

	return 0;
}
